# *** imports

# ** core
from typing import Any 

# ** app
from .request import RequestContext
from .feature import FeatureContext
from .error import ErrorContext
from ..domain import *


# *** contexts

# ** context: app_interface_context
class AppInterfaceContext(Model): 

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The application name.'
        )
    )

    # * field: features
    features = ModelType(
        FeatureContext,
        required=True,
        metadata=dict(
            description='The feature context.'
        )
    )

    # * field: errors
    errors = ModelType(
        ErrorContext,
        required=True,
        metadata=dict(
            description='The error context.'
        )
    )

    # * method: init
    def __init__(self, app_name: str, feature_context: FeatureContext, error_context: ErrorContext):
        '''
        Initialize the application interface context.

        :param app_name: The application name.
        :type app_name: str
        :param feature_context: The feature context.
        :type feature_context: FeatureContext
        :param error_context: The error context.
        :type error_context: ErrorContext
        '''

        # Initialize the model.
        super().__init__(dict(
            name=app_name,
            features=feature_context,
            errors=error_context
        ))

    # * method: parse_request
    def parse_request(self, request: Any, **kwargs) -> RequestContext:
        '''
        Parse the incoming request.

        :param request: The incoming request.
        :type request: Any
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The request context.
        :rtype: RequestContext
        '''

        # Parse request.
        return request
    
    # * method: execute_feature
    def execute_feature(self, request: RequestContext, **kwargs):
        '''
        Execute the feature context.

        :param request: The request context.
        :type request: RequestContext
        '''

        # Execute feature context and return session.
        self.features.execute(request, **kwargs)
    
    # * method: handle_response
    def handle_response(self, request: RequestContext) -> Any:
        '''
        Handle the response.

        :param request: The request context.
        :type request: RequestContext
        :return: The response.
        :rtype: Any
        '''
        
        # Import the JSON module.
        import json

        # Return the response.
        return json.loads(request.result)
    
    # * method: run
    def run(self, **kwargs):
        '''
        Run the application interface.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        
        # Parse request.
        request = self.parse_request(**kwargs)

        # Execute feature context and return session.
        # Handle error and return response if triggered.
        has_error, message = self.errors.handle_error(lambda: self.execute_feature(request, **kwargs))

        # Handle error if present.
        if has_error:
            return message

        # Handle response.
        return self.handle_response(request)




from schematics.types.serializable import serializable

from ..domain import *
from ..domain.feature import Feature, FeatureCommand


class FeatureCommandData(FeatureCommand, DataObject):
    '''
    A data representation of a feature handler.
    '''

    class Options():
        '''
        The default options for the feature handler data.
        '''

        # Set the serialize when none flag to false.
        serialize_when_none = False

        # Define the roles for the feature handler data.
        roles = {
            'to_object.yaml': DataObject.allow(),
            'to_data.yaml': DataObject.allow()
        }

    def map(self, role: str = 'to_object', **kwargs) -> FeatureCommand:
        '''
        Maps the feature handler data to a feature handler object.
        
        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature handler object.
        :rtype: f.FeatureCommand
        '''
        return super().map(FeatureCommand, role, **kwargs)


class FeatureData(Feature, DataObject):
    '''
    A data representation of a feature.
    '''

    class Options():
        '''
        The default options for the feature data.
        '''

        # Set the serialize when none flag to false.
        serialize_when_none = False

        # Define the roles for the feature data.
        roles = {
            'to_object.yaml': DataObject.deny('feature_key'),
            'to_data.yaml': DataObject.deny('feature_key', 'group_id', 'id')
        }

    commands = t.ListType(t.ModelType(FeatureCommandData),
                          deserialize_from=['handlers', 'functions', 'commands'],)
    
    @serializable
    def feature_key(self):
        '''
        Gets the feature key.
        '''

        # Return the feature key.
        return self.id.split('.')[-1]

    def map(self, role: str = 'to_object.yaml', **kwargs) -> Feature:
        '''
        Maps the feature data to a feature object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature object.
        :rtype: f.Feature
        '''

        # Map the feature data to a feature object.
        return super().map(Feature, role, **kwargs)

    @staticmethod
    def new(**kwargs) -> 'FeatureData':
        '''
        Initializes a new FeatureData object from a Feature object.
        
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureData object.
        :rtype: FeatureData
        '''

        # Create a new FeatureData object.
        _data = FeatureData(
            dict(**kwargs,), 
            strict=False
        )

        # Validate and return the new FeatureData object.
        _data.validate()
        return _data

    @staticmethod
    def from_yaml_data(id: str, group_id: str, **kwargs) -> 'FeatureData':
        '''
        Initializes a new FeatureData object from yaml data.
        
        :param id: The feature id.
        :type id: str
        :param group_id: The context group id.
        :type group_id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureData object.
        :rtype: FeatureData
        '''

        # Create a new FeatureData object.
        _data = FeatureData(
            dict(**kwargs, 
                 id=id, group_id=group_id
            ), 
            strict=False)

        # Validate and return the new FeatureData object.
        _data.validate()
        return _data



# *** imports

# ** app
from ..configs import *
from ..domain import DataObject
from ..domain.error import Error, ErrorMessage


# *** data

# ** data: error_message_data
class ErrorMessageData(ErrorMessage, DataObject):
    '''
    A data representation of an error message object.
    '''

    class Options():
        serialize_when_none = False
        roles = {
            'to_data.yaml': DataObject.allow('id'),
            'to_object.yaml': DataObject.allow()
        }


# ** data: error_data
class ErrorData(Error, DataObject):
    '''
    A data representation of an error object.
    '''

    class Options():
        serialize_when_none = False
        roles = {
            'to_data.yaml': DataObject.deny('id'),
            'to_object.yaml': DataObject.allow()
        }

    # * attribute: message
    message = ListType(
        ModelType(ErrorMessageData),
        required=True,
        metadata=dict(
            description='The error messages.'
        )
    )

    # * method: map
    def map(self, role: str = 'to_object.yaml', **kwargs):
        '''
        Maps the error data to an error object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new error object.
        :rtype: Error
        '''

        # Map the error messages.
        return super().map(Error, role, **kwargs)

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'ErrorData':
        '''
        Creates a new ErrorData object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ErrorData object.
        :rtype: ErrorData
        '''

        # Create a new ErrorData object.
        return ErrorData(
            super(ErrorData, ErrorData).new(**kwargs)
        )



# *** imports

# ** app
from ..configs import *
from ..domain import *


# *** models

# ** model: app_dependency
class AppDependency(ModuleDependency):

    # * attribute: attribute_id
    attribute_id = StringType(
        required=True,
        metadata=dict(
            description='The attribute id for the application dependency.'
        ),
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppDependency':
        '''
        Initializes a new AppDependency object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppDependency object.
        :rtype: AppDependency
        '''

        # Create and return a new AppDependency object.
        return super(AppDependency, AppDependency).new(
            AppDependency,
            **kwargs,
        )

# ** model: app_interface
class AppInterface(Entity):
    '''
    The base application interface object.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the application interface.'
        ),
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the application interface.'
        ),
    )

    # attribute: feature_flag
    feature_flag = StringType(
        required=True,
        default='core',
        metadata=dict(
            description='The feature flag.'
        ),
    )

    # attribute: data_flag
    data_flag = StringType(
        required=True,
        metadata=dict(
            description='The data flag.'
        ),
    )

    # * attribute: dependencies
    dependencies = ListType(
        ModelType(AppDependency),
        required=True,
        default=[],
        metadata=dict(
            description='The application interface dependencies.'
        ),
    )

    # * attribute: constants
    constants = DictType(
        StringType,
        default=dict(
            container_config_file='app/configs/container.yml',
            feature_config_file='app/configs/features.yml',
            error_config_file='app/configs/errors.yml',
        ),
        metadata=dict(
            description='The application dependency constants.'
        ),
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppInterface':
        '''
        Initializes a new AppInterface object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppInterface object.
        :rtype: AppInterface
        '''

        # Create and return a new AppInterface object.
        return super(AppInterface, AppInterface).new(
            AppInterface,
            **kwargs
        )
    
    # * method: get_dependency
    def get_dependency(self, attribute_id: str) -> AppDependency:
        '''
        Get the dependency by attribute id.

        :param attribute_id: The attribute id of the dependency.
        :type attribute_id: str
        :return: The dependency.
        :rtype: AppDependency
        '''

        # Get the dependency by attribute id.
        return next((dep for dep in self.dependencies if dep.attribute_id == attribute_id), None)
    

# ** model: app_repository_configuration
class AppRepositoryConfiguration(ModuleDependency):
    '''
    The import configuration for the application repository.
    '''

    # * attribute: module_path
    module_path = StringType(
        required=True,
        default='tiferet.repos.app',
        metadata=dict(
            description='The module path for the application repository.'
        ),
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        default='YamlProxy',
        metadata=dict(
            description='The class name for the application repository.'
        ),
    )

    # * attribute: params
    params = DictType(
        StringType,
        default=dict(
            app_config_file='app/configs/app.yml',
        ),
        metadata=dict(
            description='The application repository configuration parameters.'
        ),
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppRepositoryConfiguration':
        '''
        Initializes a new AppRepositoryConfiguration object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppRepositoryConfiguration object.
        :rtype: AppRepositoryConfiguration
        '''

        # Create and return a new AppRepositoryConfiguration object.
        return AppRepositoryConfiguration(
            super(AppRepositoryConfiguration, AppRepositoryConfiguration).new(
                AppRepositoryConfiguration,
                **kwargs),
            strict=False,
        )


# ** model: app_configuration
class AppConfiguration(Entity):
    '''
    The application configuration object.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the application.'
        )
    )

    # * attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the application.'
        )
    )

    # * attribute: app_repo
    app_repo = ModelType(AppRepositoryConfiguration,
        required=True,
        metadata=dict(
            description='The application repository configuration.'
        ),
    )

    # * attribute: interfaces
    interfaces = ListType(
        ModelType(AppInterface),
        required=True,
        default=[],
        metadata=dict(
            description='The application interfaces.'
        )
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppConfiguration':
        '''
        Initializes a new AppConfiguration object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppConfiguration object.
        :rtype: AppConfiguration
        '''

        # Create and return a new AppConfiguration object.
        return super(AppConfiguration, AppConfiguration).new(
            AppConfiguration,
            **kwargs
        )


# *** imports

# ** core
from typing import Any

# ** infra
from dependencies import Injector


# *** functions

# ** function: import_dependency
def import_dependency(module_path: str, class_name: str, **kwargs) -> Any:
    '''
    Import an object dependency from its configured Python module.

    :param module_path: The module path.
    :type module_path: str
    :param class_name: The class name.
    :type class_name: str
    :param kwargs: Additional keyword arguments.
    :type kwargs: dict
    :return: The dependency.
    :rtype: Any
    '''

    # Import module.
    from importlib import import_module
    return getattr(import_module(module_path), class_name)


# ** function: create_injector
def create_injector(name: str, **dependencies) -> Any:
    '''
    Create an injector object with the given dependencies.

    :param name: The name of the injector.
    :type name: str
    :param dependencies: The dependencies.
    :type dependencies: dict
    :return: The injector object.
    :rtype: Any
    '''

    # Create container.
    return type(f'{name.capitalize()}Container', (Injector,), {**dependencies})



# *** imports

# ** core
from typing import Dict

# ** app
from ..configs import *
from ..domain import DataObject
from ..domain.app import AppDependency, AppInterface, AppRepositoryConfiguration


# *** data

# ** data: app_dependency_yaml_data
class AppDependencyYamlData(AppDependency, DataObject):
    '''
    A YAML data representation of an app dependency object.
    '''

    # * attribute: attribute_id
    attribute_id = StringType(
        metadata=dict(
            description='The attribute id for the application dependency.'
        ),
    )

    class Options():
        '''
        The options for the app dependency data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.allow(),
            'to_data.yaml': DataObject.deny('attribute_id')
        }

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppDependencyYamlData':
        '''
        Initializes a new YAML representation of an AppDependency object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppDependencyData object.
        :rtype: AppDependencyData
        '''

        # Create a new AppDependencyData object.
        return AppDependencyYamlData(
            dict(**kwargs),
            strict=False,
        )
    
    # * method: map
    def map(self, **kwargs) -> AppDependency:
        '''
        Maps the app dependency data to an app dependency object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app dependency object.
        :rtype: AppDependency
        '''

        # Map the app dependency data.
        return super().map(AppDependency, **kwargs)
    

# ** data: app_interface_yaml_data
class AppInterfaceYamlData(AppInterface, DataObject):
    '''
    A data representation of an app interface object.
    '''

    class Options():
        '''
        The options for the app interface data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('app_context', 'container_context', 'feature_context', 'error_context', 'feature_repo', 'container_repo', 'error_repo'),
            'to_data.yaml': DataObject.deny('id')
        }

    # attribute: app_context
    app_context = ModelType(
        AppDependencyYamlData,
        required=True,
        metadata=dict(
            description='The application context dependency.'
        ),
    )

    # * attribute: feature_context
    feature_context = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='feature_context',
            module_path='tiferet.contexts.feature',
            class_name='FeatureContext',
        ),
        metadata=dict(
            description='The feature context dependency.'
        ),
    )

    # * attribute: container_context
    container_context = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='container_context',
            module_path='tiferet.contexts.container',
            class_name='ContainerContext',
        ),
        metadata=dict(
            description='The container context dependency.'
        ),
    )

    # * attribute: error_context
    error_context = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='error_context',
            module_path='tiferet.contexts.error',
            class_name='ErrorContext',
        ),
        metadata=dict(
            description='The error context dependency.'
        ),
    )

    # * attribute: feature_repo
    feature_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='feature_repo',
            module_path='tiferet.repos.feature',
            class_name='YamlProxy',
        ),
        metadata=dict(
            description='The feature repository dependency.'
        ),
    )

    # * attribute: container_repo
    container_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='container_repo',
            module_path='tiferet.repos.container',
            class_name='YamlProxy',
        ),
        metadata=dict(
            description='The container repository dependency.'
        ),
    )

    # * attribute: error_repo
    error_repo = ModelType(
        AppDependencyYamlData,
        required=True,
        default=AppDependencyYamlData.new(
            attribute_id='error_repo',
            module_path='tiferet.repos.error',
            class_name='YamlProxy',
        ),
        metadata=dict(
            description='The error repository dependency.'
        ),
    )

    # * method: new
    @staticmethod
    def new(app_context: Dict[str, str],
            container_context: Dict[str, str] = None,
            feature_context: Dict[str, str] = None,
            error_context: Dict[str, str] = None,
            feature_repo: Dict[str, str] = None,
            container_repo: Dict[str, str] = None,
            error_repo: Dict[str, str] = None,
            **kwargs) -> 'AppInterfaceYamlData':
        '''
        Initializes a new YAML representation of an AppInterface object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppInterfaceData object.
        :rtype: AppInterfaceData
        '''

        # Format the dependencies.
        dependencies = {}
        if app_context:
            dependencies['app_context'] = AppDependencyYamlData.new(attribute_id='app_context', **app_context)
        if container_context:
            dependencies['container_context'] = AppDependencyYamlData.new(attribute_id='container_context', **container_context)
        if feature_context:
            dependencies['feature_context'] = AppDependencyYamlData.new(attribute_id='feature_context', **feature_context)
        if error_context:
            dependencies['error_context'] = AppDependencyYamlData.new(attribute_id='error_context', **error_context)
        if feature_repo:
            dependencies['feature_repo'] = AppDependencyYamlData.new(attribute_id='feature_repo', **feature_repo)
        if container_repo:
            dependencies['container_repo'] = AppDependencyYamlData.new(attribute_id='container_repo', **container_repo)
        if error_repo:
            dependencies['error_repo'] = AppDependencyYamlData.new(attribute_id='error_repo', **error_repo)

        # Create a new AppInterfaceData object.
        data = AppInterfaceYamlData(dict(
            **dependencies,
            **kwargs),
            strict=False,
        )

        # Validate and return the new AppInterfaceData object.
        data.validate()
        return data
    
    # * method: map
    def map(self, **kwargs) -> AppInterface:
        '''
        Maps the app interface data to an app interface object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app interface object.
        :rtype: AppInterface
        '''

        # Format and map the dependencies.
        dependencies = [
            self.app_context.map(),
            self.container_context.map(),
            self.feature_context.map(),
            self.error_context.map(),
            self.feature_repo.map(),
            self.container_repo.map(),
            self.error_repo.map(),
        ]

        # Map the app interface data.
        return super().map(AppInterface, 
            dependencies=dependencies, 
            **self.to_primitive('to_model'), 
            **kwargs
        )


# ** data: app_repository_configuration_yaml_data
class AppRepositoryConfigurationYamlData(DataObject):
    '''
    A YAML data representation of an app repository configuration object.
    '''

    class Options():
        '''
        The options for the app repository configuration data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.allow(),
            'to_data.yaml': DataObject.allow()
        }

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'AppRepositoryConfigurationYamlData':
        '''
        Initializes a new YAML representation of an AppRepositoryConfiguration object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new AppRepositoryConfigurationData object.
        :rtype: AppRepositoryConfigurationData
        '''

        # Create a new AppRepositoryConfigurationData object.
        return AppRepositoryConfigurationYamlData(
            super(AppRepositoryConfigurationYamlData, AppRepositoryConfigurationYamlData).new(
                **kwargs
            )
        )
    
    # * method: map
    def map(self, **kwargs) -> AppRepositoryConfiguration:
        '''
        Maps the app repository configuration data to an app repository configuration object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app repository configuration object.
        :rtype: AppRepositoryConfiguration
        '''

        # Map the app repository configuration data.
        return super().map(AppRepositoryConfiguration, **kwargs)


# *** imports

# ** core
from typing import Any

# ** app
from ..domain import *
from ..services import container_service
from ..repos.container import ContainerRepository


# *** contexts

# ** contexts: container_context
class ContainerContext(Model):
    '''
    A container context is a class that is used to create a container object.
    '''

    # * attribute: interface_id
    interface_id = StringType(
        required=True,
        metadata=dict(
            description='The interface ID.'
        ),
    )

    # * attribute: attributes
    attributes = DictType(
        ModelType(ContainerAttribute), 
        default={}, 
        required=True,
        metadata=dict(
            description='The container attributes.'
        ),
    )

    # * attribute: constants
    constants = DictType(
        StringType, 
        default={},
        metadata=dict(
            description='The container constants.'
        ),
    )

    # * attribute: feature_flag
    feature_flag = StringType(
        required=True,
        default='core',
        metadata=dict(
            description='The feature flag.'
        ),
    )

    # * attribute: data_flag
    data_flag = StringType(
        required=True,
        metadata=dict(
            description='The data flag.'
        ),
    )

    # * method: init
    def __init__(self, interface_id: str, container_repo: ContainerRepository, feature_flag: str, data_flag: str):
        '''
        Initialize the container context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param container_repo: The container repository.
        :type container_repo: ContainerRepository
        :param interface_flag: The interface flag.
        :type interface_flag: str
        :param feature_flag: The feature flag.
        :type feature_flag: str 
        :param data_flag: The data flag.
        :type data_flag: str
        '''

        # Add the attributes as an empty dictionary.
        attributes = {}
        
        # Get and set attributes and constants.
        attrs, consts = container_repo.list_all()
        
        # Add the attributes to the context.
        for attr in attrs:
            attribute: ContainerAttribute = attributes[attr.id]

            # If the attribute already exists, set the dependencies.
            if attr.id in attributes:
                for dep in attr.dependencies:
                    attribute.set_dependency(dep)
                    continue

            # Otherwise, add the attribute.
            attributes[attr.id] = attr

        # Add the constants and attributes to the context.
        super().__init__(dict(
            interface_id=interface_id,
            feature_flag=feature_flag,
            data_flag=data_flag,
            attributes=attributes,
            constants=consts,
        ))

    # * method: get_dependency
    def get_dependency(self, attribute_id: str):
        '''
        Get a dependency from the container.

        :param attribute_id: The attribute id of the dependency.
        :type attribute_id: str
        :return: The attribute value.
        :rtype: Any
        '''

        # Create the injector.
        injector = self.create_injector()

        # Get attribute.
        return getattr(injector, attribute_id)

    # * method: create_injector
    def create_injector(self, **kwargs) -> Any:
        '''
        Add a container to the context.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The container injector object.
        :rtype: Any
        '''

        # Import dependencies.
        dependencies = {}
        for attribute_id in self.attributes:
            attribute = self.attributes[attribute_id]
            flag_map = dict(
                feature=self.feature_flag,
                data=self.data_flag,
            )
            dependencies[attribute_id] = self.import_dependency(attribute, flag_map[attribute.type])

        # Create container.
        return container_service.create_injector(self, 
            self.interface_id, 
            **self.constants, 
            **dependencies, 
            **kwargs)

    # * method: import_dependency
    def import_dependency(self, attribute: ContainerAttribute, flag: str) -> Any:
        '''
        Import a container attribute dependency from its configured Python module.

        :param attribute: The container attribute.
        :type attribute: ContainerAttribute
        :param flag: The flag for the dependency.
        :type flag: str
        :return: The dependency.
        :rtype: Any
        '''

        # Get the dependency.
        dependency = attribute.get_dependency(flag)

        # If there is no dependency and the attribute is a feature, import the default feature.
        if not dependency and attribute.type == 'feature':
            dependency = attribute.get_dependency('core')

        # Import the dependency.
        return container_service.import_dependency(dependency.module_path, dependency.class_name)


# *** imports

# ** core
from typing import Dict

# ** app
from ..domain import *


# *** contexts

# ** context: request_context
class RequestContext(Model):
    '''
    The context for an application request.
    '''

    # * attribute: feature_id
    feature_id = StringType(
        required=True,
        metadata=dict(
            description='The feature identifier for the request.'
        )
    )

    # * attribute: headers
    headers = DictType(
        StringType(),
        metadata=dict(
            description='The request headers.'
        )
    )

    # * attribute: data
    data = DictType(
        StringType(),
        metadata=dict(
            description='The request data.'
        )
    )

    # * attribute: result
    result = StringType(
        metadata=dict(
            description='The request result.'
        )
    )

    # * method: init
    def __init__(self, feature_id: str, headers: Dict[str, str], data: Dict[str, str]):
        '''
        Initialize the request context object.

        :param feature_id: The feature identifier.
        :type feature_id: str
        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Set the context attributes.
        self.feature_id = feature_id
        self.headers = headers
        self.data = data

        # Validate the context.
        self.validate()

    # * method: set_result
    def set_result(self, result: Any):
        '''
        Set the serialized result value.

        :param result: The result object.
        :type result: Any
        '''

        # Import the json module.
        import json

        # Set the result as a serialized empty dictionary if it is None.
        if not result:
            self.result = json.dumps({})
            return
            
        # If the result is a Model, convert it to a primitive dictionary and serialize it.
        if isinstance(result, Model):
            self.result = json.dumps(result.to_primitive())
            return

        # If the result is not a list, it must be a dict, so serialize it and set it.
        if type(self.result) != list:
            self.result = json.dumps(result)
            return

        # If the result is a list, convert each item to a primitive dictionary.
        result_list = []
        for item in result:
            if isinstance(item, Model):
                result_list.append(item.to_primitive())
            else:
                result_list.append(item)

        # Serialize the result and set it.
        self.result = json.dumps(result_list)



from ..domain.feature import Feature
from ..domain.feature import FeatureHandler
from ..repos.feature import FeatureRepository


class AddNewFeature(object):
    '''
    Add a new feature.
    '''

    def __init__(self, feature_repo: FeatureRepository):
        '''
        Initialize the command.
        
        :param feature_repo: The feature repository.
        :type feature_repo: FeatureRepository
        '''

        # Set the feature repository.
        self.feature_repo = feature_repo

    def execute(self, **kwargs) -> Feature:
        '''
        Execute the command to add a new feature.
        
        :param kwargs: The keyword arguments.
        :type kwargs: dict
        :return: The new feature.
        '''

        # Create a new feature.
        feature = Feature.new(**kwargs)

        # Assert that the feature does not already exist.
        assert not self.feature_repo.exists(
            feature.id), f'FEATURE_ALREADY_EXISTS: {feature.id}'

        # Save and return the feature.
        self.feature_repo.save(feature)
        return feature


class AddFeatureHandler(object):
    '''
    Adds a feature handler to a feature.
    '''

    def __init__(self, feature_repo: FeatureRepository):
        '''
        Initialize the command.
        
        :param feature_repo: The feature repository.
        :type feature_repo: FeatureRepository
        '''

        # Set the feature repository.
        self.feature_repo = feature_repo

    def execute(self, feature_id: str, position: int = None, **kwargs):
        '''
        Execute the command to add a feature handler to a feature.

        :param feature_id: The feature ID.
        :type feature_id: str
        :param position: The position of the handler.
        :type position: int
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The updated feature.
        :rtype: Feature
        '''

        # Create a new feature handler instance.
        handler = FeatureHandler.new(**kwargs)

        # Get the feature using the feature ID.
        feature = self.feature_repo.get(feature_id)

        # Assert that the feature was successfully found.
        assert feature is not None, f'FEATURE_NOT_FOUND: {feature_id}'

        # Add the feature handler to the feature.
        feature.add_handler(
            handler,
            position=position
        )

        # Save and return the feature.
        self.feature_repo.save(feature)
        return feature



# *** imports

# ** infra
from schematics import types as t


# *** config

# ** config (class): string_type
class StringType(t.StringType):
    '''
    A string type.
    '''

    pass


# ** config (class): integer_type
class IntegerType(t.IntType):
    '''
    An integer type.
    '''

    pass


# ** config (class): float_type
class FloatType(t.FloatType):
    '''
    A float type.
    '''

    pass


# ** config (class): boolean_type
class BooleanType(t.BooleanType):
    '''
    A boolean type.
    '''

    pass


# ** config (class): list_type
class ListType(t.ListType):
    '''
    A list type.
    '''

    pass


# ** config (class): dict_type
class DictType(t.DictType):
    '''
    A dictionary type.
    '''

    pass


# ** config (class): model_type
class ModelType(t.ModelType):
    '''
    A model type.
    '''

    pass



# *** imports

# ** core
from typing import List

# ** app
from ..data.feature import FeatureData
from ..domain.feature import Feature
from ..clients import yaml_client

# *** repository

# ** interface: feature_repository
class FeatureRepository(object):
    '''
    Feature repository interface.
    '''

    # * method: exists
    def exists(self, id: str) -> bool:
        '''
        Verifies if the feature exists.

        :param id: The feature id.
        :type id: str
        :return: Whether the feature exists.
        :rtype: bool
        '''

        # Not implemented.
        raise NotImplementedError()

    # * method: get
    def get(self, id: str) -> Feature:
        '''
        Get the feature by id.

        :param id: The feature id.
        :type id: str
        :return: The feature object.
        :rtype: f.Feature
        '''

        # Not implemented.
        raise NotImplementedError()

    # * method: list
    def list(self, group_id: str = None) -> List[Feature]:
        '''
        List the features.

        :param group_id: The group id.
        :type group_id: str
        :return: The list of features.
        :rtype: list
        '''

        # Not implemented.
        raise NotImplementedError

    # * method: save
    def save(self, feature: Feature):
        '''
        Save the feature.
        
        :param feature: The feature object.
        :type feature: f.Feature
        '''

        # Not implemented.
        raise NotImplementedError()

# ** repository: yaml_proxy
class YamlProxy(FeatureRepository):
    '''
    Yaml repository for features.
    '''

    # * method: init
    def __init__(self, feature_config_file: str):
        '''
        Initialize the yaml repository.

        :param feature_config_file: The feature configuration file.
        :type feature_config_file: str
        '''

        # Set the base path.
        self.config_file = feature_config_file

    # * method: exists
    def exists(self, id: str) -> bool:
        '''
        Verifies if the feature exists.
        
        :param id: The feature id.
        :type id: str
        :return: Whether the feature exists.
        :rtype: bool
        '''

        # Retrieve the feature by id.
        feature = self.get(id)

        # Return whether the feature exists.
        return feature is not None

    # * method: get
    def get(self, id: str) -> Feature:
        '''
        Get the feature by id.
        
        :param id: The feature id.
        :type id: str
        :return: The feature object.
        '''

        # Get context group and feature key from the id.
        group_id, feature_key = id.split('.')

        # Load feature data from yaml.
        _data: FeatureData = yaml_client.load(
            self.config_file,
            create_data=lambda data: FeatureData.from_yaml_data(
                id=id,
                group_id=group_id,
                **data
            ),
            start_node=lambda data: data.get('features').get('groups').get(group_id).get('features').get(feature_key)
        )

        # Return None if feature data is not found.
        if not _data:
            return None

        # Return feature.
        return _data.map('to_object.yaml')
    
    # * method: list
    def list(self, group_id: str = None) -> List[Feature]:
        '''
        List the features.
        
        :param group_id: The group id.
        :type group_id: str
        :return: The list of features.
        :rtype: list
        '''

        # Load all feature data from yaml.
        features = yaml_client.load(
            self.config_file,
            create_data=lambda data: [FeatureData.from_yaml_data(
                id=id,
                **feature_data
            ) for id, feature_data in data.items()],
            start_node=lambda data: data.get('features')
        )

        # Filter features by group id.
        if group_id:
            features = [feature for feature in features if feature.group_id == group_id]

        # Return the list of features.
        return [feature.map('to_object.yaml') for feature in features]

    # * method: save
    def save(self, feature: Feature):
        '''
        Save the feature.
        
        :param feature: The feature object.
        :type feature: f.Feature
        '''

        # Create updated feature data.
        feature_data = FeatureData.new(**feature.to_primitive())

        # Update the feature data.
        yaml_client.save(
            self.config_file,
            data=feature_data,
            data_save_path=f'features/{feature.group_id}.{feature_data.feature_key}'
        )

        # Return the updated feature object.
        return feature_data.map('to_object.yaml')



# *** imports

# ** core
from typing import Any, Tuple

# ** app
from ..domain import *
from ..repos.error import ErrorRepository


# *** contexts

# ** context: error_context
class ErrorContext(Model):
    '''
    The error context object.
    '''

    # * attribute: errors
    errors = DictType(
        ModelType(Error),
        required=True,
        metadata=dict(
            description='The errors lookup.'
        )
    )

    # * method: init
    def __init__(self, error_repo: ErrorRepository):
        '''
        Initialize the error context object.
        
        :param error_repo: The error repository.
        :type error_repo: ErrorRepository
        '''

        # Create the errors lookup from the error repository.
        errors = {error.name: error for error in error_repo.list()}

        # Set the errors lookup and validate.
        super().__init__(dict(errors=errors))
        self.validate()

    # * method: handle_error
    def handle_error(self, execute_feature: Any) -> Tuple[bool, Any]:
        '''
        Handle an error.

        :param func: The execute feature function to handle.
        :type func: function
        :return: Whether the error was handled.
        :rtype: bool
        '''

        # Execute the feature function and handle the errors.
        try:
            execute_feature()
            return (False, None)
        except AssertionError as e:
            return (True, self.format_error_response(str(e)))

    # * method: format_error_response
    def format_error_response(self, error_message: str, lang: str = 'en_US', **kwargs) -> Any:
        '''
        Format the error response.

        :param error_message: The error message.
        :type error_message: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The formatted error message.
        :rtype: Any
        '''

        # Split error message.
        try:
            error_name, error_data = error_message.split(': ')
        except ValueError:
            error_name = error_message
            error_data = None

        # Format error data if present.
        error_data = error_data.split(', ') if error_data else None

        # Get error.
        error = self.errors.get(error_name)

        # Set error response.
        error_response = dict(
            message=error.format(lang, *error_data if error_data else []),
            **kwargs
        )

        # Return error response.
        return error_response



# *** imports

# ** app
from ..domain import *


# *** models

# ** model: feature_command
class FeatureCommand(ValueObject):
    '''
    A command object for a feature command.
    '''

    # * attribute: name
    name = t.StringType(
        required=True,
        metadata=dict(
            description='The name of the feature handler.'
        )
    )

    # * attribute: attribute_id
    attribute_id = t.StringType(
        required=True,
        metadata=dict(
            description='The container attribute ID for the feature command.'
        )
    )

    # * attribute: params
    params = t.DictType(
        t.StringType(),
        default={},
        metadata=dict(
            description='The custom parameters for the feature handler.'
        )
    )

    # * attribute: return_to_data
    return_to_data = t.BooleanType(
        metadata=dict(
            description='Whether to return the feature command result to the feature data context.'
        )
    )

    # * attribute: data_key
    data_key = t.StringType(
        metadata=dict(
            description='The data key to store the feature command result in if Return to Data is True.'
        )
    )

    # * attribute: pass_on_error
    pass_on_error = t.BooleanType(
        metadata=dict(
            description='Whether to pass on the error if the feature handler fails.'
        )
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'FeatureCommand':
        '''Initializes a new FeatureCommand object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureCommand object.
        '''

        # Create a new FeatureCommand object.
        obj = FeatureCommand(dict(
            **kwargs
        ), strict=False)

        # Validate and return the new FeatureCommand object.
        obj.validate()
        return obj


# ** model: feature
class Feature(Entity):
    '''
    A feature object.
    '''

    # * attribute: name
    name = t.StringType(
        required=True,
        metadata=dict(
            description='The name of the feature.'
        )
    )

    # * attribute: group_id
    group_id = t.StringType(
        required=True,
        metadata=dict(
            description='The context group identifier for the feature.'
        )
    )

    # * attribute: description
    description = t.StringType(
        metadata=dict(
            description='The description of the feature.'
        )
    )

    # * attribute: commands
    commands = t.ListType(
        t.ModelType(FeatureCommand),
        default=[],
        metadata=dict(
            description='The command handler workflow for the feature.'
        )
    )

    # * attribute: log_params
    log_params = t.DictType(
        t.StringType(),
        default={},
        metadata=dict(
            description='The parameters to log for the feature.'
        )
    )

    # * method: new
    @staticmethod
    def new(name: str, group_id: str, feature_key: str, description: str = None, **kwargs) -> 'Feature':
        '''Initializes a new Feature object.

        :param name: The name of the feature.
        :type name: str
        :param group_id: The context group identifier of the feature.
        :type group_id: str
        :param feature_key: The key of the feature.
        :type feature_key: str
        :param description: The description of the feature.
        :type description: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new Feature object.
        '''

        # Feature ID is the group ID and feature key separated by a period.
        id = f'{group_id}.{feature_key}'

        # Set the description as the name if not provided.
        if not description:
            description = name

        # Create and return a new Feature object.
        return super(Feature, Feature).new(
            id=id,
            name=name,
            group_id=group_id,
            description=description,
            **kwargs
        )
    
    # * method: add_handler
    def add_handler(self, handler: FeatureCommand, position: int = None):
        '''Adds a handler to the feature.

        :param handler: The handler to add.
        :type handler: FeatureCommand
        :param position: The position to add the handler at.
        :type position: int
        '''

        # Add the handler to the feature.
        if position:
            self.commands.insert(position, handler)
        else:
            self.commands.append(handler)


# *** imports

# ** app
from ..domain.app import AppRepositoryConfiguration


# *** constants

# ** constant: APP_REPO
APP_REPO = AppRepositoryConfiguration.new(
    module_path='tiferet.repos.app',
    class_name='YamlProxy',
    params=dict(
        app_config_file='app/configs/app.yaml'
    ),
)


# *** imports

# ** infra
from schematics import Model

# ** app
from .app import AppInterfaceContext
from ..services import container_service
from ..domain import *
from ..repos.app import AppRepository


# *** contexts

# ** context: environment_context
class EnvironmentContext(Model):
    '''
    An environment context is a class that is used to create and run the app interface context.
    '''

    # * attribute: interfaces
    interfaces = DictType(
        ModelType(AppInterface), 
        default={},
        metadata=dict(
            description='The app interfaces keyed by interface ID.'
        ),
    )

    # * method: init
    def __init__(self, **kwargs):
        '''
        Initialize the environment context.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Load the app repository.
        app_repo = self.load_app_repo()

        # Load the interface configuration.
        interfaces = {interface.id: interface for interface in app_repo.list_interfaces()}

        # Set the interfaces.
        super().__init__(dict(
            interfaces=interfaces,
            **kwargs
        ), strict=False)

    # * method: start
    def start(self, interface_id: str, **kwargs):
        '''
        Start the environment context.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Load the app context.
        app_context = self.load_app_context(interface_id)

        # Run the app context.
        app_context.run(
            interface_id=interface_id,
            **kwargs
        )

    # * method: load_app_repo
    def load_app_repo(self) -> AppRepository:
        '''
        Load the app interface repository.

        :return: The app repository.
        :rtype: AppRepository
        '''

        # Load the app repository configuration.
        from ..configs.app import APP_REPO

        # Return the app repository.
        return container_service.import_dependency(APP_REPO.module_path, APP_REPO.class_name)(**APP_REPO.params)

    # * method: load_app_context
    def load_app_context(self, interface_id: str) -> AppInterfaceContext:
        '''
        Load the app context.

        :param container: The app container.
        :type container: AppContainer
        :return: The app context.
        :rtype: AppContext
        '''

        # Get the app interface.
        app_interface: AppInterface = self.interfaces.get(interface_id)

        # Get the default dependencies for the app interface.
        dependencies = dict(
            interface_id=app_interface.id,
            app_name=app_interface.name,
            feature_flag=app_interface.feature_flag,
            data_flag=app_interface.data_flag,
            app_context=container_service.import_dependency(
                **app_interface.app_context.to_primitive()
            ),
            **app_interface.constants
        )

        # Import the dependencies.
        for dep in app_interface.dependencies:
            dependencies[dep.attribute_id] = container_service.import_dependency(dep.module_path, dep.class_name)

        # Create the injector from the dependencies, constants, and the app interface.
        injector = container_service.create_injector(
            app_interface.id,
            **dependencies
        )

        # Return the app context.
        return getattr(injector, 'app_context')
